#!/usr/bin/perl

# Try to keep the programming style good ;-)

use strict;
use warnings;

# Some modules we need for password-related Stuff

use Crypt::PasswdMD5;
use Crypt::RandPasswd;

# We want to parse commandline arguments in a comfortable way

use Getopt::Std;

# Parse the commandline arguments and fall back to default values if
# neccessary

my %opts;
getopts('b:g:i:l:S:o:m:x:e:D:t:hrsdy', \%opts);

if ( $opts{h} ) { usage(); }

sub usage { die "usage: $0 [-i input file] [-g group] [-b base directory for \$HOME] [-s] [-S smbpasswd binary] [-o output file] [-r] [-m minimum password length] [-x maximum password length] [-d] [-D <descriptors for the fields in input file (semicolon separated)> [-y]] [-t <perl regexp that describes the delimiter in the input file>]\n";}

if ( $> != 0 ) {die "You must be root to run this.\n";}

if ( ! scalar %opts ) { usage(); }

my $userliste = "userlist";
if ( $opts{i} ) { $userliste = $opts{i}; }

my $group = "schueler";
if ( $opts{g} ) { $group = $opts{g}; }

if ( ! getgrnam $group ) { die "Group $group doesn't exist. Please create it or specify an existing group.\n" }

my $homebase = "/home/$group";
if ( $opts{b} ) { $homebase = $opts{b}; }

my $smbpass = "/usr/local/bin/smbpasswd";
if ( $opts{S} ) { $smbpass = $opts{S}; }

my $outfile = "useraccounts";
if ( $opts{o} ) { $outfile = $opts{o}; }

my $passtype = "letters";
if ( $opts{r} ) { $passtype = "chars"; }

my @descriptor = ("username","real name","class");
if ( $opts{D} ) { @descriptor = split(/\;/, $opts{D}); }

my $delimiter = '/\t/';
if ( $opts{t} ) { $delimiter = $opts{t}; }

if ( $delimiter !~ /^\/.*\/$/ ) { die "$delimiter is not a valid regular expression. A correct example would be '/\t/'.\n"; }

# Although we require the person running the script to supply the
# enclosing slashes of the regexp we remove them now as the pattern
# won't work otherwise.

$delimiter =~ s/^\/(.*)\/$/$1/;

my $minlength;
if ( $opts{m} ) { $minlength = $opts{m}; }
else	{
	if ( $opts{r} ) { $minlength = 10; }
	else { $minlength = 8; }
	}

my $maxlength;
if ( $opts{x} ) { $maxlength = $opts{x}; }
else	{
	if ( $opts{r} ) { $maxlength = 10; }
	else { $maxlength = 8; }
	}

print "Group: $group\n";
print "Base directory: $homebase\n";
print "Userliste: $userliste\n\n";

# Determine the system we run on

my $uname;
if ( ! stat "/usr/bin/uname" ) { $uname = "/bin/uname"; }
else { $uname = "/usr/bin/uname" }

my $os = `$uname`;
chomp $os;

# Create the filehandle to read the userlist from or read from STDIN if
# $opts{i} is "-"

my $handle;
if ( $opts{i} eq "-" ) { $handle ="STDIN"; }
else
	{
	$handle = "LISTE";
	open LISTE,$userliste or die "$userliste: No such file or directory.\n";
	}

# Do some deep magic to determine the school year.

# First get the raw time
my @time = localtime;

# Extract the relevant parts

my $month = $time[4];
my $year = $time[5];

# Convert the year to 2-digit format

my $year1 = sprintf("%02d", $year % 100);

# Now it's time for the arcane part...

my $schoolyear;
if ($month le 7) {
	my $year2=$year1 - 1;
	$schoolyear = "0$year2/$year1";
}
else {
	my $year2=$year1 + 1;
	$schoolyear = "$year1/0$year2";
}

# Make -s implicit if -S is given

if ( $opts{S} and ! $opts{s} )
	{
	$opts{s} = 1;
	}

# Check for smbpasswd(8) if -s is given

if ( $opts{s} ) 
	{
	if ( ! -x $smbpass ) { die "$smbpass doesn't exist or isn't executable. Please specify.\n"; }
	}

# Extract the account data from the list of usernames

while (my $line = <$handle>)
{
	# Parse the userlist and put it's elements in variables

	my @entry = split(/$delimiter/, $line);

	# First check if @descriptor and @entry contain an equal number
	# of entries

	if ( (scalar @descriptor) != (scalar @entry) ) { die "The number of fields in the input file doesn't match the number of descriptors\n"; }

	my @descriptor_process = @descriptor;

	# shorten @descriptor by one to match @entry
	shift @descriptor_process;
	
	my $name = shift @entry;

	# Remove the newline left over from parsing the input file
	chomp @entry;

	my $output_rest;
	my @entry_process = @entry;
	while ( @entry_process && @descriptor_process)
		{
		$output_rest .= (shift @descriptor_process) . ": " . (shift @entry_process ). " ";
		}
	if ( $opts{y} ) { $output_rest .= $schoolyear };

	$output_rest =~ s/\;$//;

	# Determine if we should try appending random characters to duplicate
	# user names or simply log and report them.

	if ( $opts{d} ) { dupe_engine($name); }
	else { next if dupe_log($name); }

	sub dupe_engine
		{
		# Check if the user exists
		my $username = $_[0];
		if ( getpwnam $username ) 
			{
			print "User $username already in passwd database.\n";
			unless ( ! getpwnam $username )
				{
				my $randomizer = Crypt::RandPasswd->word(1,2);
				$username .= $randomizer;
				print "New Name: $username\n";
				}
			}
		}
	sub dupe_log
		{
		# Check if the user exists
		my $username = $_[0];
		if ( getpwnam $username )
			{
			my $errlog = "duplicate_users";
			if ( $opts{e} ) { $errlog = $opts{e}; }
			print "User $username already in passwd database.\n";
			open DUPELOG,">> $errlog";
			print DUPELOG "Username: $username $output_rest\n";
			close DUPELOG;
			}
		else {return;}
		}
	
	# Compose the GECOS field for /etc/passwd

	my $gecos;
	while ( @entry )
		{
		$gecos .= (shift @entry) . ";";
		}

	$gecos =~ s/\;$//;


	# Generate a password for the user

	my $clearpass = Crypt::RandPasswd->$passtype( $minlength, $maxlength );

	my $passwdhash = unix_md5_crypt ($clearpass, '$1$');

	# Create an Unix account
	# First determine which sytem we run on and which commands we
	# have to use

	if ( $os eq "Linux" )
		{
		# useradd(8) is braindead because it won't let me
		# specify a base directory for $HOME
		my $home = "$homebase/$name";
		if ( $home =~ /\/{2,}/ ) { $home =~ s/\/{2,}/\//g; }

		system "/usr/sbin/useradd -m -c '$gecos' -g $group -d $home $name";
		open PASSPIPE,"| /usr/sbin/chpasswd -e";
		print PASSPIPE "$name:$passwdhash\n";
		close PASSPIPE;
		}

	if ( $os eq "FreeBSD" )
		{
		open PWPIPE, "| /usr/sbin/pw useradd -n $name -m -c '$gecos' -g $group -b $homebase -h 0";
		print PWPIPE "$clearpass\n";
		close PWPIPE;
		}

	if ( $os eq "NetBSD" )
		{
		die "None of the NetBSD user management tools will accept the\npassword on standard input and it would have to be supplied as\ncommandline argument. If you really want to do this\nuncomment line 141 and comment line 140 out I warned you.\n";
#		system "/usr/sbin/useradd -m -c '$gecos' -g $group -b $homebase -p $passwdhash $name";
		}

	# Use smbpasswd(8) to create an smbpasswd entry
	# Yes this is ugly...and so is smbpasswd's non-interactive
	# mode...

	# Compose the complete smbpasswd command

	if ( $opts{s} ) { smb_create(); }

	sub smb_create 
		{
		my $smbpasswd = "$smbpass -s -a $name";
		open SMBPIPE,"| $smbpasswd";
		print SMBPIPE "$clearpass\n$clearpass\n";
		close SMBPIPE;
		}

	# Write the account data to a file

	open OUTFILE,">>$outfile";
	print OUTFILE "Username: $name password: $clearpass $output_rest\n";
	print "Username: $name password: $clearpass $output_rest\n";
	
}


close LISTE;
close OUTFILE;

chmod (0600, $outfile);
